#!/usr/bin/perl

# Convert a raw binary blob into a .cas file.

use IO::Handle;

sub suggest_help {
	print STDERR "Try '$0 --help' for more information'\n";
	exit 0;
}

sub help_text {
	print STDERR <<EOF;
usage: $0 OPTION... input-file
       $0 --blocks TYPE:LENGTH[,TYPE:LENGTH]... input-file
Generate a cassette image from binary input.

  -D                   input file is a DragonDOS binary
  -C                   input file is a CoCo RSDOS binary
  -l ADDR              load address in filename block
  -e ADDR              exec address in filename block
  -n NAME              name in filename block
  -o FILE              output file (defaults to stdout)
      --leader COUNT   leader size before filename block and data
      --no-filename    no filename block required in output
      --no-eof         no EOF block required in output
      --eof-data       EOF block allowed to contain the last chunk of data

In standard cassette image mode, load and exec addresses are determined from
the binary if DragonDOS or RSDOS binary format is specified, otherwise they
should be specified as an option.

With --blocks, each block TYPE and LENGTH is explicitly specified.  If LENGTH
is negative, this means "use all but the last -LENGTH bytes".  If TYPE is the
character "l", a leader of LENGTH bytes is emitted.
EOF
	exit 0;
}

my $load = 1024;
my $exec = -1;
my $name;
my $want_fnblock = 1;
my $want_eofblock = 1;
my $eof_data = 0;
my $leader = 256;
my $mode = '';
my @blocks = ();

while (my $opt = shift @ARGV) {
	if ($opt eq '--') {
		last;
	} elsif ($opt eq '-D') {
		$mode = 'dragon';
	} elsif ($opt eq '-C') {
		$mode = 'coco';
	} elsif ($opt eq '-l') {
		my $v = shift @ARGV;
		if ($v =~ /^(\d+|0x[\da-f]+)$/i) {
			$load = eval $1;
		}
	} elsif ($opt eq '-e') {
		my $v = shift @ARGV;
		if ($v =~ /^(\d+|0x[\da-f]+)$/i) {
			$exec = eval $1;
		}
	} elsif ($opt eq '-n') {
		$name = shift @ARGV;
	} elsif ($opt eq '-o') {
		my $f = shift @ARGV;
		my $o;
		open($o, ">", $f) or die $!;
		STDOUT->fdopen(\*$o, 'w') or die $!;
	} elsif ($opt eq '--leader') {
		my $v = shift @ARGV;
		if ($v =~ /^(\d+|0x[\da-f]+)$/i) {
			$leader = eval $1;
		}
	} elsif ($opt eq '--no-filename') {
		$want_fnblock = 0;
	} elsif ($opt eq '--no-eof') {
		$want_eofblock = 0;
	} elsif ($opt eq '--eof-data') {
		$eof_data = 1;
	} elsif ($opt eq '--blocks') {
		my $v = shift @ARGV;
		push @blocks, split(/,/, $v);
	} elsif ($opt eq '--help') {
		help_text();
	} elsif ($opt =~ /^-/) {
		print STDERR "Unrecognised option '$opt'\n";
		suggest_help();
	} else {
		unshift @ARGV, $opt;
		last;
	}
}

die "no input file\n" unless scalar(@ARGV) > 0;
my $file = shift @ARGV;
my $size = (-s $file);

my $in;
open($in, "<", $file) or die "failed to open $file: $!\n";
binmode $in;
binmode STDOUT;

# Special case: if a list of blocks is specified, always treat the file
# as raw data, and output blocks only as specified.

if (scalar(@blocks) > 0) {
	my $type = 1;
	while ($block = shift @blocks) {
		if ($block eq 'l') {
			print "U" x $leader;
		} elsif ($block =~ /^(\d+|0x[\da-f]+)(:(-?\d+|0x[\da-f]+))?$/) {
			my $type = eval $1;
			my $bsize;
			if (defined $3) {
				$bsize = eval $3;
				$bsize += $size if ($bsize < 0);
				$bsize = 255 if ($bsize > 255);
			} else {
				$bsize = ($size > 255) ? 255 : $size;
			}
			my $data;
			my $length = sysread $in, $data, $bsize;
			$size -= $bsize;
			block_out(STDOUT, $type, $data);
		} else {
			last;
		}
	}
	exit 0;
}

# Othewise, a standard cassette image is required.

if (!defined $name) {
	if ($file =~ /^(\w{1,8})\./) {
		$name = uc $1;
	} else {
		$name = "BINARY";
	}
}

my $filetype = 2;  # binary

if ($mode eq 'dragon') {
	getc($in);  # skip $55
	$filetype = getc($in);
	$load = (unpack("C", getc($in)) << 8) | unpack("C", getc($in));
	my $length = (unpack("C", getc($in)) << 8) | unpack("C", getc($in));
	$exec = (unpack("C", getc($in)) << 8) | unpack("C", getc($in));
	getc($in);  # skip $55
	$size -= 9;
	if ($size != $length) {
		print STDERR "Warning: file size mismatch - using shorter value\n";
	}
	if ($length < $size) {
		$size = $length;
	}
} elsif ($mode eq 'coco') {
	getc($in);  # skip $00
	my $length = (unpack("C", getc($in)) << 8) | unpack("C", getc($in));
	$load = (unpack("C", getc($in)) << 8) | unpack("C", getc($in));
	if ($exec < 0 && ($length + 10) <= $size) {
		# assume the next chunk will be the exec address
		seek $in, $length + 8, 0;
		$exec = (unpack("C", getc($in)) << 8) | unpack("C", getc($in));
		# seek back
		seek $in, 5, 0;
		$size -= 5;
	}
	$size -= 5;
	if ($size != $length) {
		print STDERR "Warning: file size mismatch - using shorter value\n";
	}
	if ($length < $size) {
		$size = $length;
	}
}

$exec = $load if ($exec < 0);

if ($want_fnblock) {
	print "U" x $leader;
	my $data = sprintf("%-8s%c\x00\x00%c%c%c%c", $name, $filetype, $exec>>8, $exec & 0xff, $load >> 8, $load & 0xff);
	block_out(STDOUT, 0, $data);
}

print "U" x $leader;

while ($size > 0) {
	my $data;
	my $bsize = ($size > 255) ? 255 : $size;
	my $length = sysread $in, $data, $bsize;
	if ($length != $bsize) {
		print STDERR "Warning, short read on last block\n";
	}
	$size -= $bsize;
	if ($size <= 0 && $want_eofblock) {
		if ($eof_data) {
			block_out(STDOUT, 0xff, $data);
		} else {
			block_out(STDOUT, 1, $data);
			block_out(STDOUT, 0xff, "");
		}
	} else {
		block_out(STDOUT, 1, $data);
	}
}

print "U";
exit 0;

sub block_out {
	my ($f,$type,$data) = @_;
	my $length = length($data);
	my $sum = $type + $length;;
	print $f pack("C*", 0x55, 0x3c, $type, length($data));
	print $f $data;
	for (unpack("C*", $data)) {
		$sum += $_;
	}
	print $f pack("C*", ($sum & 0xff), 0x55);
}
