#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

/* simple zip.  encodes as:
 * 1iiiiiii 0nnnnnnn - repeat 128-n (1-128) bytes from current + 1x (-ve 2s cmp)
 * 1iiiiiii 1jjjjjjj nnnnnnnn - repeat 128-n (1-256) bytes from current + 1ij
 * 0nnnnnnn - directly copy next 128-n (1-128) bytes */

/* mmap()ed file */
static uint8_t *map;

/* cursor position within the data map */
static off_t map_pos;

/* amount of bytes remaining */
static off_t bytes_remaining;

/* buffer for stdin */
static uint8_t *stdin_buffer;

/* Best copy offset/length found */
static int copy_offset, copy_length;

/* Current run */
static off_t run_start;
int run_length;

static void dump_run(void);
static void dump_copy(void);
static int count_copy_length(off_t from, off_t to);
static void find_copy(off_t to);

static void usage(const char *prog) {
	fprintf(stderr, "Usage: %s [-c] [infile]\n", prog);
}

/*
static void missing_arg(const char *opt) {
	fprintf(stderr, "Missing argument to option %s\n", opt);
	exit(1);
}
*/

static int to_stdout = 0;
static int from_stdin = 0;
static int have_eof = 0;

int main(int argc, char **argv) {
	int argn;
	int in;

	/* parse options */
	argn = 1;
	while (argn < argc && argv[argn][0] == '-') {
		if (0 == strcmp(argv[argn], "-c")) {
			to_stdout = 1;
		} else {
			usage(argv[0]);
			exit(1);
		}
		argn++;
	}

	char *in_fname = NULL;
	if (argn < argc) {
		in_fname = argv[argn++];
		in = open(in_fname, O_RDONLY);
		if (in < 0) {
			fprintf(stderr, "Failed to open %s for reading: ", in_fname);
			perror(NULL);
			exit(1);
		}

		struct stat statbuf;
		if (fstat(in, &statbuf) < 0) {
			perror(NULL);
			exit(1);
		}
		if (!(statbuf.st_mode & S_IFREG)) {
			fprintf(stderr, "%s is not a regular file\n", in_fname);
			exit(1);
		}
		bytes_remaining = statbuf.st_size;

		map = mmap(NULL, bytes_remaining, PROT_READ, MAP_PRIVATE, in, 0);
		if (map == MAP_FAILED) {
			perror("mmap failed");
			exit(1);
		}
	} else {
		to_stdout = 1;
		from_stdin = 1;
		stdin_buffer = malloc(0x8100);
		if (!stdin_buffer) {
			perror("malloc failed");
			exit(1);
		}
		map = stdin_buffer;
		bytes_remaining = fread(stdin_buffer, 1, 0x8100, stdin);
		if (bytes_remaining < 0x8100) {
			if (!feof(stdin)) {
				perror("short read");
				exit(1);
			}
			have_eof = 1;
		}
	}

	if (!to_stdout) {
		char *out_fname = malloc(strlen(in_fname) + 4);
		if (!out_fname) {
			perror("malloc() failed");
			exit(1);
		}
		strcpy(out_fname, in_fname);
		strcat(out_fname, ".dz");
		if (freopen(out_fname, "wb", stdout) == NULL) {
			fprintf(stderr, "Failed to open '%s' for writing: ", out_fname);
			perror(NULL);
			exit(1);
		}
	}

	run_start = run_length = 0;
	copy_offset = copy_length = 0;
	map_pos = 0;
	while (bytes_remaining > 0) {
		if (from_stdin && !have_eof && map_pos >= 0x8000) {
			memmove(stdin_buffer, stdin_buffer + 0x4000, 0x4100);
			size_t nread = fread(stdin_buffer + 0x4100, 1, 0x4000, stdin);
			if (nread < 0x4000) {
				if (!feof(stdin)) {
					perror("short read");
					exit(1);
				}
				have_eof = 1;
			}
			bytes_remaining += nread;
			map_pos -= 0x4000;
			if (run_start > 0)
				run_start -= 0x4000;
		}
		find_copy(map_pos);
		/* short offset run > 2 bytes always worth doing */
		if (copy_offset >= -128) {
			if (copy_length > 2
			    /* or 2 bytes, and not already in a run */
			    || (copy_length == 2 && run_length == 0)) {
				dump_copy();
				continue;
			}
		}
		/* long offset run > 3 bytes always worth doing */
		if (copy_length > 3
		    /* or 3 bytes, and not already in a run */
		    || (copy_length == 3 && run_length == 0)) {
			dump_copy();
			continue;
		}
		/* if already in a run, continue it */
		if (run_length > 0) {
			run_length++;
			map_pos++;
			bytes_remaining--;
			/* maximum run length is 128 */
			if (run_length == 128) {
				dump_run();
			}
			continue;
		}
		/* otherwise, start a run */
		run_start = map_pos;
		run_length = 1;
		map_pos++;
		bytes_remaining--;
	}
	dump_run();

	if (!to_stdout) {
		unlink(in_fname);
	}

	return 0;
}

static void dump_run(void) {
	if (run_length == 0)
		return;
	int i;
	fputc(128 - run_length, stdout);
	//fprintf(stderr, "run (%d): ", run_length);
	for (i = 0; i < run_length; i++) {
		fputc(map[run_start + i], stdout);
		//fprintf(stderr, "%02x ", map[run_start + i]);
	}
	//fprintf(stderr, "\n");
	run_length = run_start = 0;
}

static void dump_copy(void) {
	dump_run();
	if (copy_offset >= -128 && copy_length <= 128) {
		fputc(copy_offset & 0xff, stdout);
		fputc(128 - copy_length, stdout);
		//fprintf(stderr, "copy (%d): %d\n", copy_length, copy_offset);
	} else {
		fputc(((copy_offset >> 7) & 0x7f) | 0x80, stdout);
		fputc((copy_offset & 0x7f) | 0x80, stdout);
		fputc((384 - copy_length) & 0xff, stdout);
		//fprintf(stderr, "long copy (%d): %d\n", copy_length, copy_offset);
	}
	map_pos += copy_length;
	bytes_remaining -= copy_length;
}

/* check possible length of copy at 'from' */
static int count_copy_length(off_t from, off_t to) {
	int length = 0;
	while (length < 256) {
		if (length >= bytes_remaining)
			return length;
		if (map[from] != map[to])
			return length;
		from++;
		to++;
		length++;
	}
	return length;
}

/* find best copy */
static void find_copy(off_t to) {
	copy_length = 0;
	/* work backwards: even if an extra byte is used for offset (further
	 * back), a longer run exceeding the best so far will be at least as
	 * efficient */
	int f;
	for (f = 1; f <= 0x4000; f++) {
		if (f > to)
			return;
		int l = count_copy_length(to - f, to);
		if (l > copy_length) {
			copy_length = l;
			copy_offset = -f;
		}
	}
}
