#!/usr/bin/perl

# Read a MIDI file and dump out to custom music player format.
# Define sample base frequency with -s sample=midi_note.
# Map MIDI channels with -m track,inchannel=outchannel,voice (where outchannel
# is 0-2).  Define voices with -v voice=wave[,wave]...
# Specify output note position divider with -d - this divides the MIDI
# position.
# Output position range can be specified with -r start,end.

use IO::Handle;

sub suggest_help {
	print STDERR "Try '$0 --help' for more information'\n";
	exit 0;
}

sub help_text {
	print STDERR <<EOF;
usage: $0 OPTION... input-file
Parses a MIDI input file and generates data for custom music player format.

  -s VOICE=NOTE              base frequency of sampled VOICE is MIDI NOTE
  -m MTRK,MCHAN=CHAN,VOICE   maps MIDI track/channel to VOICE on output CHAN
  -v VOICE=WAVE[,WAVE]...    maps output VOICE to a list of WAVEs
  -d DIV                     MIDI times are divided by this per beat fragment
  -r START,END               range of notes (in output fragments) to include
  -c FREQ                    CPU frequency [894886.25]
  -x CYCLES                  number of cycles in mixer loop [89]
  -o FILENAME                output filename
  -l FILENAME                log filename - useful to identify MTRKs/MCHANs

The WAVEs in a VOICE are played once each per beat fragment.  Built in
VOICEs are defined as a simple decay over 6 fragments: saw, sqr, tri, sin.
Three different default noise VOICEs are defined as a shorter decay over
3 fragments: nz0, nz1, nz2.

When defining a mapping with -m, an MTRK of '_' matches all MIDI tracks.

Output data is in the form of assembly directives, designed to be included in
source and assembled.  Data is sent to stdout by default.

If two notes with the same VOICE are played on the same CHAN, data is emitted
to generate a tremolo between the two.
EOF
	exit 0;
}

use constant HEADER_CHUNK => 0x4d546864;
use constant TRACK_CHUNK  => 0x4d54726b;

my %samples = ();
my %voices = (
	saw => [ 'saw2', 'saw2', 'saw2', 'saw1', 'saw1', 'saw0' ],
	sqr => [ 'sqr2', 'sqr2', 'sqr2', 'sqr1', 'sqr1', 'sqr0' ],
	tri => [ 'tri2', 'tri2', 'tri2', 'tri1', 'tri1', 'tri0' ],
	sin => [ 'sin2', 'sin2', 'sin2', 'sin1', 'sin1', 'sin0' ],
	nz0 => [ 'nz02', 'nz22', 'nz10' ],
	nz1 => [ 'nz12', 'nz02', 'nz20' ],
	nz2 => [ 'nz22', 'nz12', 'nz00' ],
);
my %track_voice = ();
my $div = 8;
my $start_position = 0;
my $end_position = 1024;
my $log_filename;
my $cpu_freq = 14318180 / 16;
my $mixer_cyc = 89;

while (my $opt = shift @ARGV) {
	if ($opt eq '--') {
		last;
	} elsif ($opt eq '-s') {
		my $v = shift @ARGV;
		if ($v =~ /^(\w+)=(\d+)$/) {
			$samples{$1} = $2;
		}
	} elsif ($opt eq '-m') {
		my $v = shift @ARGV;
		if ($v =~ /^(\d+|_),(\d+)=(\d+),(\w+)$/) {
			$track_voice{$1}->{$2} = { channel => $3, voice => $4 };
		}
	} elsif ($opt eq '-v') {
		my $v = shift @ARGV;
		if ($v =~ /^(\w+)=([\w,]+)$/) {
			$voices{$1} = [ split /,/, $2 ];
		}
	} elsif ($opt eq '-d') {
		$div = shift @ARGV;
	} elsif ($opt eq '-r') {
		my $v = shift @ARGV;
		if ($v =~ /^(\d+),(\d+)$/) {
			$start_position = $1;
			$end_position = $2 - $1;
		}
	} elsif ($opt eq '-c') {
		my $v = shift @ARGV;
		$cpu_freq = 0 + $v;
	} elsif ($opt eq '-x') {
		my $v = shift @ARGV;
		if ($v =~ /^(\d+|0x[\da-f]+)$/) {
			$mixer_cyc = eval $1;
		}
	} elsif ($opt eq '-o') {
		my $v = shift @ARGV;
		my $o;
		open($o, ">", $v) or die $!;
		STDOUT->fdopen(\*$o, 'w') or die $!;
	} elsif ($opt eq '-l') {
		$log_filename = shift @ARGV;
	} elsif ($opt eq '--help') {
		help_text();
	} elsif ($opt =~ /^-/) {
		print STDERR "Unrecognised option '$opt'\n";
		suggest_help();
	} else {
		unshift @ARGV, $opt;
		last;
	}
}

unless (scalar(@ARGV) > 0) {
	print STDERR "No input file specified.\n";
	suggest_help();
}
my $i;
open($i, "<", $ARGV[0]) or die $!;
STDIN->fdopen(\*$i, 'r') or die $!;

if (defined $log_filename) {
	my $e;
	open($e, ">", $log_filename) or die $!;
	STDERR->fdopen(\*$e, 'w') or die $!;
} else {
	close STDERR;
}

my $chunk_type = 0;
my $chunk_bytes = 0;

my $eof = 0;

sub next_chunk {
	while ($chunk_bytes > 0) {
		byte_in();
		last if ($eof);
	}
	$chunk_type = (byte_in()<<24) | (byte_in()<<16) | (byte_in()<<8) | byte_in();
	my $tmp = (byte_in()<<24) | (byte_in()<<16) | (byte_in()<<8) | byte_in();
	$chunk_bytes = $tmp;
}

sub byte_in {
	my $b = getc();
	if (!defined $b) {
		$eof = 1;
		return 0;
	}
	$chunk_bytes--;
	return unpack("C", $b);
}

sub word_in {
	return (byte_in() << 8) | byte_in();
}

sub delta_in {
	my $size = 0;
	my $b;
	do {
		$b = byte_in();
		$size = ($size << 7) | ($b & 0x7f);
	} while (($b & 0x80) == 0x80);
	return $size;
}

my @notes = ();

next_chunk();
if ($chunk_type != HEADER_CHUNK) {
	print STDERR "No header chunk: aborting.\n";
	exit 1;
}

my $midi_track = -1;

CHUNK: while (1) {
	next_chunk();
	last CHUNK if ($eof);
	while ($chunk_type != TRACK_CHUNK) {
		next_chunk();
		last CHUNK if ($eof);
	}
	$midi_track++;
	my %note = ();
	my $midi_position = 0;
	my $position = -$start_position;  # == position / div
	COMMAND: while ($chunk_bytes > 0) {
		last CHUNK if ($eof);
		next CHUNK if ($position >= $end_position);
		my $delta = delta_in();
		my $cmd = byte_in();
		# some midi files are weird:
		#while ($cmd < 128 && $chunk_bytes > 0) {
		#	$cmd = byte_in();
		#}
		next CHUNK if ($chunk_bytes <= 0);
		$midi_position += $delta;
		$position = int($midi_position / $div) - $start_position;
		if ($cmd == 0xff) {
			$cmd = byte_in();
			my $length = byte_in();
			print STDERR "META-EVENT $midi_track: $cmd ($length)\n";
			for my $i (0 .. $length-1) {
				byte_in();
			}
			next COMMAND;
		}
		my $midi_channel = $cmd & 0xf;
		$cmd >>= 4;
		if ($cmd == 0x8) {
			# NOTE OFF
			my $nn = byte_in();
			my $vv = byte_in();
			print STDERR "NOTE OFF $midi_position ($position) $midi_track,$midi_channel: $nn\n";
			next COMMAND unless ($position >= 0);
			my $tv;
			if (exists $track_voice{$midi_track}->{$midi_channel}) {
				$tv = $track_voice{$midi_track}->{$midi_channel};
			} elsif (exists $track_voice{'_'}->{$midi_channel}) {
				$tv = $track_voice{'_'}->{$midi_channel};
			} else {
				next COMMAND;
			}
			my $out_channel = $tv->{channel};
			next COMMAND unless (exists $note{$out_channel}->{$nn});
			my $n = $note{$out_channel}->{$nn};
			my $voice = $n->{voice};
			my $start = $n->{start};
			my $duration = $position - $start;
			$note{$out_channel}->{$nn}->{n}->{duration} = $duration;
			delete $note{$out_channel}->{$nn};
		} elsif ($cmd == 0x9) {
			# NOTE ON
			my $nn = byte_in();
			my $vv = byte_in();
			print STDERR "NOTE ON $midi_position ($position) track $midi_track channel $midi_channel: $nn\n";
			next COMMAND unless ($position >= 0);
			my $tv;
			if (exists $track_voice{$midi_track}->{$midi_channel}) {
				$tv = $track_voice{$midi_track}->{$midi_channel};
			} elsif (exists $track_voice{'_'}->{$midi_channel}) {
				$tv = $track_voice{'_'}->{$midi_channel};
			} else {
				next COMMAND;
			}
			my $out_channel = $tv->{channel};
			my $voice = $tv->{voice};
			play_note($out_channel, $voice, $nn, $position, 16);
			$note{$out_channel}->{$nn} = {
				n => $notes[$out_channel]->{$position},
				start => $position,
			};
		} elsif ($cmd == 0xa) {
			my $nn = byte_in();
			my $vv = byte_in();
			print STDERR "AFTER-TOUCH $midi_track,$midi_channel: $nn $vv\n";
		} elsif ($cmd == 0xb) {
			my $cc = byte_in();
			my $vv = byte_in();
			print STDERR "CONTROL-CHANGE $midi_track,$midi_channel: $nn $vv\n";
		} elsif ($cmd == 0xc) {
			my $pp = byte_in();
			print STDERR "PATCH-CHANGE $midi_track,$midi_channel: $pp\n";
		} elsif ($cmd == 0xd) {
			my $cc = byte_in();
			print STDERR "CHANNEL AFTER-TOUCH $midi_track,$midi_channel: $cc\n";
		} elsif ($cmd == 0xe) {
			my $pc = (byte_in() & 0x7f) | (byte_in() << 7);
			print STDERR "PITCH-WHEEL-CHANGE $midi_track,$midi_channel: ".$pc-0x2000."\n";
		} else {
			print STDERR "UNKNOWN COMMAND $cmd channel $midi_channel\n";
		}
	}
}

# truncate overlapping notes
for my $c (0..2) {
	for my $p (0..$end_position-1) {
		if (exists $notes[$c]->{$p}) {
			my $test = $notes[$c]->{$p};
			if (($test->{duration} + $p) > $end_position) {
				print STDERR "TRUNCATING $c $p $test->{duration}\n";
				$test->{duration} = $end_position - $p;
			}
			TEST: for my $i (1 .. $test->{duration}-1) {
				if (exists $notes[$c]->{$p+$i}) {
					$test->{duration} = $i;
					last TEST;
				}
			}
		}
	}
}

# convert voices
for my $c (1..2) {
	for my $p (sort {$a<=>$b} keys %{$notes[$c]}) {
		my $note = $notes[$c]->{$p};
		my $voice = $note->{voice};
		my $freq = $note->{freq};
		my $duration = $note->{duration};
		my $freq0 = $freq;
		my $freq1 = $freq;
		my $duration0 = $duration;
		my $duration1 = $duration;
		if (exists $note->{freq0}) {
			$freq0 = $note->{freq0};
			$freq1 = $note->{freq1};
			$duration0 = $note->{duration0};
			$duration1 = $note->{duration1};
		}
		if (exists $voices{$voice}) {
			delete $notes[$c]->{$p};
			my $i = 0;
			my $vlist = $voices{$voice};
			while ($i < $duration && $i <= $#{$vlist}) {
				last if exists $notes[$c]->{$p+$i};
				# tremolo between two frequencies
				if (($i & 1) == 0) {
					$freq = $freq0;
				} else {
					$freq = $freq1;
				}
				if ($i >= $duration0) {
					$freq = $freq1;
				}
				if ($i >= $duration1) {
					$freq = $freq0;
				}
				$notes[$c]->{$p+$i} = {
					voice => $vlist->[$i],
					freq => $freq,
					duration => 1,
				};
				$i++;
			}
		}
	}
}

# insert silences
for my $c (0..2) {
	my $p = 0;
	my $gap_start = -1;
	PLOOP: while ($p < $end_position) {
		my $gap_duration = $p - $gap_start;
		while (exists $notes[$c]->{$p}) {
			if ($gap_start >= 0) {
				$notes[$c]->{$gap_start} = {
					voice => 'silence',
					freq => 0,
					duration => $gap_duration,
				};
				$gap_start = -1;
			}
			my $test = $notes[$c]->{$p};
			$p += $test->{duration};
			last PLOOP if ($p >= $end_position);
		}

		$gap_start = $p if ($gap_start < 0);
		$p++;
	}
	if ($gap_start >= 0) {
		my $gap_duration = $p - $gap_start;
		$notes[$c]->{$gap_start} = {
			voice => 'silence',
			freq => 0,
			duration => $gap_duration,
		};
	}
}

for my $p (0..$end_position-1) {
	my @chan_str = ();
	for my $c (0..2) {
		if (exists $notes[$c]->{$p}) {
			push @chan_str, "p_channel".($c+1);
		}
	}
	my $chan_str = join(" | ", @chan_str) || "0";
	print "\tfcb\t$chan_str\n";

	for my $c (0..2) {
		#print "\t; $p ch $c\n";
		if (exists $notes[$c]->{$p}) {
			my $note = $notes[$c]->{$p};
			my $voice = $note->{voice};
			my $freq = $note->{freq};
			my $f;
			if ($c == 0) {
				my $bnote = 69;
				if (exists $samples{$voice}) {
					$bnote = $samples{$voice};
				}
				my $bfreq = (2 ** (($bnote - 69) / 12)) * 440;
				$f = (256 * $freq) / $bfreq;
			} else {
				$f = (65536 * $freq) / ($cpu_freq / $mixer_cyc);
			}
			if ($c == 0) {
				print "\tfdb\t${voice}_start\n";
				print "\tfdb\t${voice}_end\n";
				print "\tfdb\t${voice}_restart\n";
				print "\tfdb\t".int($f)."\n";
			} else {
				print "\tfcb\t($voice >> 8)\n";
				print "\tfdb\t".int($f)."\n";
			}
		}
	}
}

sub play_note {
	my ($channel, $voice, $note, $start, $duration) = @_;
	return unless ($duration > 0);
	#$note += 12 if ($channel == 1);
	my $freq = (2 ** (($note - 69) / 12)) * 440;
	if (exists $notes[$channel]->{$start}) {
		my $n = $notes[$channel]->{$start};
		if ($n->{voice} eq $voice) {
			$n->{duration0} = $n->{duration};
			$n->{freq0} = $n->{freq};
			$n->{duration1} = $duration;
			$n->{freq1} = $freq;
			if ($duration > $n->{duration}) {
				$n->{duration} = $duration;
			}
			return;
		}
	}
	$notes[$channel]->{$start} = {
		voice => $voice,
		freq => $freq,
		duration => $duration,
	};
}
