#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <SDL_image.h>

/* SDL on MacOSX... */
#ifdef main
# undef main
#endif

/* Converts an image to Semigraphics 4 (9-colour) format.  Assumes one
 * line in input image translates directly to one line in output. */

int getpixel(SDL_Surface *surface, int x, int y);

static void usage(const char *prog) {
	fprintf(stderr, "Usage: %s [-o outfile] infile\n", prog);
}

static void missing_arg(const char *opt) {
	fprintf(stderr, "Missing argument to option %s\n", opt);
	exit(1);
}

int main(int argc, char **argv) {
	int argn;
	SDL_Surface *in;
	int ymul = 1;

	/* parse options */
	argn = 1;
	while (argn < argc && argv[argn][0] == '-') {
		if (0 == strcmp(argv[argn], "-o")) {
			/* -o output filename */
			argn++;
			if (argn >= argc) missing_arg(argv[argn]);
			if (freopen(argv[argn], "wb", stdout) == NULL) {
				fprintf(stderr, "Couldn't open '%s' for writing: ", argv[argn]);
				perror(NULL);
				exit(1);
			}
		} else if (0 == strcmp(argv[argn], "-ymul")) {
			/* -ymul y-axis multiplier */
			argn++;
			if (argn >= argc) missing_arg(argv[argn]);
			ymul = atoi(argv[argn]);
		} else {
			usage(argv[0]);
			exit(1);
		}
		argn++;
	}
	/* filename must be specified */
	if (argn >= argc) {
		usage(argv[0]);
		exit(1);
	}

	in = IMG_Load(argv[argn]);
	if (!in) {
		fprintf(stderr, "Failed to open %s: %s\n", argv[argn], IMG_GetError());
		return 1;
	}
	argn++;

	int y;
	for (y = 0; y < in->h; y++) {
		int yy;
		for (yy = 0; yy < ymul; yy++) {
			int  x;
			for (x = 0; (x+1) < in->w; x += 2) {
				int byte = 128;
				int byte_colour = 0;
				int colour;

				/* for black pixels, don't care what colour the
				 * byte is, for others set the byte colour */

				colour = getpixel(in, x, y);
				if (colour > 0) {
					byte_colour = (colour - 1) << 4;
					/* in SG4 terms, set the top-left and
					 * bottom-left pixels */
					byte |= 0x0a;
				}

				colour = getpixel(in, x+1, y);
				if (colour > 0) {
					byte_colour = (colour - 1) << 4;
					/* set the top-right and bottom-right
					 * pixels */
					byte |= 0x05;
				}

				fputc(byte | byte_colour, stdout);
			}
		}
	}

	SDL_FreeSurface(in);

	return 0;
}

/* black, blue, green, cyan, red, magenta, yellow, white, orange */
static int colour_map[9] = {
	0, 3, 1, 6, 4, 7, 2, 5, 8
};

/* returns an SG4 colour from colour_map */
int getpixel(SDL_Surface *surface, int x, int y) {
	/* main pixel fetch cribbed from SDL examples */
	int bpp = surface->format->BytesPerPixel;
	Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;
	Uint32 dp;
	switch(bpp) {
		case 1:
			dp = *p;
			break;
		case 2:
			dp = *(Uint16 *)p;
			break;
		case 3:
			if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
				dp = p[0] << 16 | p[1] << 8 | p[2];
			else
				dp = p[0] | p[1] << 8 | p[2] << 16;
			break;
		case 4:
			dp = *(Uint32 *)p;
			break;
		default:
			dp = 0;
			break;
	}
	Uint8 dr, dg, db;
	SDL_GetRGB(dp, surface->format, &dr, &dg, &db);
	/* now convert that into a 3-bit RGB colour */
	int colour = (dr >= 128) ? 4 : 0;
	colour |= (dg >= 128) ? 2 : 0;
	colour |= (db >= 128) ? 1 : 0;
	/* special case for orange: */
	if (colour == 6 && dg < 224) colour = 8;
	/* index into colour_map: */
	return colour_map[colour];
}
